/*
 * Cheops Next Generation GUI
 * 
 * gui-canvas.c
 * Functions for the gnome canvas and the page_object's that are on the canvas
 *
 * Copyright(C) 1999 Brent Priddy <toopriddy@mailcity.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA
 */

#include <gnome.h>
#include <netdb.h>
#include "cheops-gui.h"
#include "gui-viewspace.h"
#include "gui-canvas.h"
#include "gui-utils.h"
#include "gui-settings.h"
#include "gui-config.h"
#include "gui-service.h"
#include "cheops-osscan.h"
#include "ip_utils.h"
#include "logger.h"
#include "gui-dns.h"
#include "gui-pixmap.h"
#include "gui-handlers.h"
#include "gui-dns.h"

//#define DEBUG_CANVAS

#ifdef DEBUG_CANVAS
	#define DEBUG(a) a
#else
	#define DEBUG(a)
#endif

typedef struct _poo_info {
	GtkWidget   *menu;
	page_object *po;
	net_page    *np;
} poo_info;


int page_object_event_dispatch(GtkObject *widget, GdkEvent *event, page_object *po);

double mouse_down_x, mouse_down_y;

GList *page_object_option_list;

extern int control_key;

void deleting_group(GtkWidget *w, page_object *po)
{
	po->group = NULL;
}

/*
 * Create a new page_abject with all fields avaliable for use
 */
page_object *page_object_new(GnomeCanvas *canvas)
{
	page_object *po;
	
	g_return_val_if_fail( canvas, NULL);
	g_return_val_if_fail( po = malloc(sizeof(page_object)), NULL );

	memset(po, 0, sizeof(page_object));

	po->flags = PAGE_OBJECT_CREATED_NEW;
	po->icon_file_name = makestring("unknown.xpm");
	po->canvas = canvas;
//	po->tooltips = gtk_tooltips_new();
//	gtk_tooltips_set_delay(po->tooltips, tooltips_timeout);
	
	po->group = gnome_canvas_item_new( gnome_canvas_root(canvas),
					  GNOME_TYPE_CANVAS_GROUP, NULL );
	po->box =   gnome_canvas_item_new( GNOME_CANVAS_GROUP(po->group),
					  GNOME_TYPE_CANVAS_RECT,
					  "fill_color", "white",
					  "outline_color", "black",
					  NULL );
	po->label = gnome_canvas_item_new( GNOME_CANVAS_GROUP(po->group),
					  GNOME_TYPE_CANVAS_TEXT,
					  "font", "fixed",
					  "fill_color", "black",
					  "anchor", GTK_ANCHOR_N,
					  NULL );
	po->icon  = gnome_canvas_item_new( GNOME_CANVAS_GROUP(po->group),
					  gnome_canvas_image_get_type(),
					  "anchor", GTK_ANCHOR_S,
					  NULL );

	// put a signal on the goup so we dont try to free it twice
	gtk_signal_connect(GTK_OBJECT(po->group), "destroy", deleting_group, po);

	gtk_signal_connect( GTK_OBJECT(po->group), "event", GTK_SIGNAL_FUNC(page_object_event_dispatch), po);

	return(po);
}

#define INITIAL_Y 50.0
#define INITIAL_X 40.0

char *page_object_get_name(page_object *po)
{
	char *name;
	if(po)
	{
		if(po->name == NULL)
			name = ip2str(po->ip);
		else
		{
			if(options_use_ip_for_label)
				name = ip2str(po->ip);
			else
				name = po->name;
		}
	}
	else
		name = "Unknown";
	return(name);
}	

/*
 * Display the page object, give it a pixmap, and put it on the canvas
 */
void page_object_display( page_object *po )
{
	double x1, x2, y1, y2;
	double nextx = 0.0;
	double nexty = INITIAL_Y;
	double height, width;
	double nextydown = 0.0;
	char *name;
	
	g_return_if_fail( po );

	// correct for the restore from config (the ip address is used as the name)
	if(po->name)
	{
		struct in_addr address;
		if(inet_aton(po->name, &address))
		{
			if(po->ip == address.s_addr)
			{
				free(po->name);
				po->name = NULL;
			}
		}
	}
/*
 * add the text and the text box (which is 2 units on each side fatter than the
 * text	is)
 */
	name = page_object_get_name(po);
	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->label), "text", name, NULL);
	gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(po->label), &x1, &y1, &x2, &y2);

	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->box), "x1", x1-2.0, "x2", x2+2.0, "y1", y1-2.0, "y2", y2+2.0, NULL);

/*
 * get rid of the image, cause we are going to redisplay it
 */ 
	if(po->im)
	{
		gdk_imlib_kill_image(po->im);
		po->im = NULL;
	}

	if(po->os_data && po->os_data->os)
	{
		if(po->icon_file_name)
			free(po->icon_file_name);
		
		po->icon_file_name = makestring(os_pixmap_list_get(po->os_data->os));	
	}
	po->im = get_my_imlib_image(po->icon_file_name);

	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->icon), "image", po->im, "width", (double)po->im->rgb_width, "height", (double)po->im->rgb_height, NULL);

//	gtk_tooltips_set_tip(po->tooltips, GTK_WIDGET(main_window->notebook), "this is a tip", NULL);
		
/*
 * if the page_object is new then place it next to the last created page_object
 */
	if(po->flags & PAGE_OBJECT_CREATED_NEW)
	{
		gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(po->group), &x1, &y1, &x2, &y2);
		height = y2 - y1;
		width = x2 - x1;
	
		if(po->next != NULL)
		{
			nextx = po->next->x + width;
			nexty = po->next->y;
			nextydown = height;
		}
		po->x = INITIAL_X + nextx;
		po->y = nexty;
	
		if((po->x + width) >= CANVAS_FIXED_WIDTH)
		{
			po->x = INITIAL_X;
			po->y += INITIAL_Y + nextydown;
		}
		if((po->y + height) >= CANVAS_FIXED_HEIGHT)
		{
			po->x = INITIAL_X;
			po->y = INITIAL_Y;
		}
		po->flags &= ~PAGE_OBJECT_CREATED_NEW;

	}
	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->group), "x", (double)po->x, "y", (double)po->y, NULL); 	
	
/*
 * This bit of code will fix the time when the border of the page_object
 * extends off of the side of the screen
 */
	gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(po->group), &x1, &y1, &x2, &y2);
	DEBUG(printf("bounds:\nx1 = %f\ny1 = %f\nx2 = %f\ny2 = %f\n",x1, y1, x2, y2));

	if(x1 < 0)
		po->x -= x1;
	if(x2 > CANVAS_FIXED_WIDTH)
		po->x -= x2 - CANVAS_FIXED_WIDTH;
	if(y1 < 0)
		po->y -= y1;
	if(y2 > CANVAS_FIXED_HEIGHT)
		po->y -= y2 - CANVAS_FIXED_HEIGHT;

	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->group), "x", po->x, "y", po->y, NULL); 	
}

page_object_link *page_object_link_new()
{
	page_object_link *pol = malloc(sizeof(*pol));
	
	if(pol)
	{
		pol->link = NULL;
		pol->po1 = NULL;
		pol->po2 = NULL;
		pol->pts = gnome_canvas_points_new(2);
	}
	return(pol);
}

void page_object_link_destroy(page_object_link *pol, page_object *po)
{
	if(pol)
	{
		if(!pol->link || 
		   !pol->po1 ||
		   !pol->po2 ||
		   !pol->pts)
		{
			g_error("oops... something bad has happened with this page_object_link");
		}
		else
		{
			/*
			 * take the GList node out of the other page_object (cause we delete the 
			 * po->link list after the g_list_foreach()  )
			 */
			if(po != pol->po1)
				pol->po1->link = g_list_remove(pol->po1->link, pol);

			if(po != pol->po2)
				pol->po2->link = g_list_remove(pol->po2->link, pol);

			gnome_canvas_points_free(pol->pts);
			gtk_object_destroy(GTK_OBJECT(pol->link));
			
			free(pol);
		}
	}	
}

/*
 * Function used to see if there is already a connection fomr one page_object
 * to another page_object (it returns the page_object_link)
 */
page_object_link *page_object_link_get(page_object *po1, page_object *po2)
{
	GList *gl;
	page_object_link *pol = NULL;
	
	if(po1 && po2 && po1->link && po2->link)
	{
		for(gl = po1->link; gl; gl = gl->next) // we only need to traverse one list
		{
			pol = gl->data;
			if( (pol->po1 == po1 && pol->po2 == po2) || (pol->po1 == po2 && pol->po2 == po1) )
			    break;
			pol = NULL;
		}
	}
	return(pol);
}

	
void net_page_map_all(net_page *np)
{
	page_object *po;
	
	for(po = np->page_objects; po; po = po->next)
		do_map_icmp(primary_agent, po->ip, np);
}

void page_object_map(net_page *np, page_object *po)
{
	if(is_valid_net_page(np))
		if(is_valid_page_object(np, po))
			do_map_icmp(primary_agent, po->ip, np);
}

void net_page_map_plink(GtkWidget *w, gpointer data)
{
	net_page *np = data;
	page_object *po;
	
	if(is_valid_net_page(np))
		for(po = np->plink; po; po = po->plink)
			do_map_icmp(primary_agent, po->ip, np);
}

void net_page_osscan_plink(GtkWidget *w, gpointer data)
{
	net_page *np = data;
	page_object *po;

	if(is_valid_net_page(np))
	{
		for(po = np->plink; po; po = po->plink)
		{
			do_os_scan(primary_agent, po->ip, np, 0);
		}
	}
}

void net_page_probe_plink(GtkWidget *w, gpointer data)
{
	net_page *np = data;
	page_object *po;
	os_port_entry *port;
	int i;
	
	if(is_valid_net_page(np))
	{
		for(po = np->plink; po; po = po->plink)
		{
			if(po->os_data)
			{
				for(port = po->os_data->ports; port; port = port->next)
				{
					do_probe(primary_agent, po->ip, port->port, probe_timeout_ms, np);
				}
			}
			else
			{
				for(i = 0; i < 100; i++)
				{
					do_probe(primary_agent, po->ip, i, probe_timeout_ms, np);
				}
			}
		}
	}
}

void net_page_dns_plink(GtkWidget *w, gpointer data)
{
	net_page *np = data;
	page_object *po;
	
	if(is_valid_net_page(np))
		for(po = np->plink; po; po = po->plink)
			dns_reverse_lookup( inet_ntoa(*(struct in_addr *)&(po->ip)), reverse_lookup_cb, NULL, po);
}

void net_page_map_page_objects(net_page *np, page_object *po, page_object *po2)
{
	page_object_link *pol;
	double x1, x2, y1, y2;
	
	if(np && po && po2 &&                                                // are our pointers ok?
	   is_valid_page_object(np, po) && is_valid_page_object(np, po2) &&  // are they valid page_objects?
	   po != po2)                                                        // dont add link if they are the same
	{
		if((pol = page_object_link_get(po, po2)) )
			page_object_link_destroy(pol, NULL);
			
		pol = page_object_link_new();

		gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(po->group), &x1, &y1, &x2, &y2);
		pol->pts->coords[0] = x1 + (x2 - x1)/2.0;
		pol->pts->coords[1] = y1 + (y2 - y1)/2.0;

		gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(po2->group), &x1, &y1, &x2, &y2);
		pol->pts->coords[2] = x1 + (x2 - x1)/2.0;
		pol->pts->coords[3] = y1 + (y2 - y1)/2.0;

		pol->link = gnome_canvas_item_new( gnome_canvas_root(GNOME_CANVAS(np->canvas)),
		                                   GNOME_TYPE_CANVAS_LINE,
		                                   "points", pol->pts,
		                                   "fill_color", "black",
		                                   "width_units", 1.0,
		                                   NULL);
		pol->po1 = po;
		pol->po2 = po2;
		
		po->link =  g_list_append(po->link, pol);
		po2->link = g_list_append(po2->link, pol);
		gnome_canvas_item_lower_to_bottom(pol->link);
	}
}

void net_page_unmap_all(net_page *np)
{
	page_object *po;

	if(np)
	{
		for(po = np->page_objects; po; po = po->next)
		{
			g_list_foreach(po->link, (GFunc)page_object_link_destroy, po);
			g_list_free(po->link);
			po->link = NULL;
		}
	}
}
	

void page_object_link_move(page_object_link *pol, page_object *po)
{
	double x1, x2, y1, y2;
	
	if(pol)
	{
		if( pol->po1 == po)
		{
			gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(po->group), &x1, &y1, &x2, &y2);
			pol->pts->coords[0] = x1 + (x2 - x1)/2.0;
			pol->pts->coords[1] = y1 + (y2 - y1)/2.0;
			DEBUG(printf("link bounds:\nx1 = %f\ny1 = %f\nx2 = %f\ny2 = %f\n",x1, y1, x2, y2));
		}
		else
		{
			if(pol->po2 == po)
			{
				gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(po->group), &x1, &y1, &x2, &y2);
				pol->pts->coords[2] = x1 + (x2 - x1)/2.0;
				pol->pts->coords[3] = y1 + (y2 - y1)/2.0;
				DEBUG(printf("link bounds:\nx1 = %f\ny1 = %f\nx2 = %f\ny2 = %f\n",x1, y1, x2, y2));
			}
			else
			{
				if(po == NULL)
				{
					gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(pol->po1->group), &x1, &y1, &x2, &y2);
					pol->pts->coords[2] = x1 + (x2 - x1)/2.0;
					pol->pts->coords[3] = y1 + (y2 - y1)/2.0;
					gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(pol->po2->group), &x1, &y1, &x2, &y2);
					pol->pts->coords[0] = x1 + (x2 - x1)/2.0;
					pol->pts->coords[1] = y1 + (y2 - y1)/2.0;
				}
				else
					printf("%s(): oops, i got an invalid page_object to move\n", __FUNCTION__);
			}
		}
		gnome_canvas_item_set( GNOME_CANVAS_ITEM(pol->link), "points", pol->pts, NULL);
	}
}

void page_object_links_move(GList *list, page_object *po)
{
	if(list)
		g_list_foreach(list, (GFunc)page_object_link_move,po);
}

void page_object_highlight_highlight(page_object *po)
{
	po->flags |= PAGE_OBJECT_HIGHLIGHT;
	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->box), "fill_color", "blue", NULL);
	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->label), "fill_color", "white", NULL);
}

void page_object_highlight_unhighlight(page_object *po)
{
	static unsigned int unhighlight = ~PAGE_OBJECT_HIGHLIGHT;

	po->flags &= unhighlight; 
	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->box), "fill_color", "white", NULL);
	gnome_canvas_item_set(GNOME_CANVAS_ITEM(po->label), "fill_color", "black", NULL);
}	

/*
 * Highlight all or unhighlight all (i know it looks nasty but it is fast)
 */
void net_page_highlight_all(net_page *np, int highlight)
{
	page_object *po;
	
	if(highlight) /* highlight all */
	{
		for(po = np->page_objects;po;po = po->next)
		{
			if(!(po->flags & PAGE_OBJECT_HIGHLIGHT))
			{
				page_object_highlight_highlight(po);
			}
		}
	}
	else
	{
		for(po = np->page_objects;po;po = po->next)
		{
			if(po->flags & PAGE_OBJECT_HIGHLIGHT)
			{
				page_object_highlight_unhighlight(po);
			}
		}
	}
}

/*
 * Toggle the page_object's highlight state
 */
void page_object_highlight_toggle(page_object *po)
{
	if(po->flags & PAGE_OBJECT_HIGHLIGHT)
		page_object_highlight_highlight(po);
	else
		page_object_highlight_unhighlight(po);
}

/*
 * remove all plink's on the current net_page
 */
void net_page_unset_plink()
{
	net_page *np = get_current_net_page();
	
	np->plink = NULL;
}

/*
 * Set up the plink adding all highlighted items on the current net_page
 */
void net_page_set_plink()
{
	net_page *np = get_current_net_page();
	page_object *po;
	
	net_page_unset_plink();
	
	for(po = np->page_objects; po; po = po->next)
	{
		if(po->flags & PAGE_OBJECT_HIGHLIGHT)
		{
			po->plink = np->plink;
			np->plink = po;
		}
	}
}

/*
 * get a page object on the net_page
 */
page_object *page_object_get(net_page *np, char *name)
{
	page_object *p;
	
	if(is_valid_net_page(np))
	{
		for(p = np->page_objects; p; p = p->next)
		{
			if(!strcmp(name, p->name))
			{
				return(p);
			}
		}
	}
	return(NULL);
}

/*
 * get a page object on the net_page by the ip address
 */
page_object *page_object_get_by_ip(net_page *np, unsigned int ip)
{
	page_object *p;
	
	if(is_valid_net_page(np))
	{
		for(p = np->page_objects; p; p = p->next)
		{
			if(p->ip == ip)
			{
				return(p);
			}
		}
	}
	return(NULL);
}

int is_valid_page_object(net_page *np, page_object *po)
{
	page_object *p;
	
	if(is_valid_net_page(np))
	{
		for(p = np->page_objects; p; p = p->next)
		{
			if(p == po)
			{
				return(TRUE);
			}
		}
	}
	return(FALSE);
}

/*
 * delete a single page_object on the net_page's canvas
 */
void page_object_delete(net_page *np, page_object *po)
{
	page_object *p, *prev = NULL;
	os_stats *os;
	os_port_entry *port;
	
	for(p = np->page_objects; p; p = p->next)
	{
		if((void *)po == (void *)p)
		{
			if(p->group)
				gtk_object_destroy(GTK_OBJECT(p->group));
			if(p->data)	
				free(p->data);
			if(p->notes)	
				free(p->notes);
			if(p->name)	
				free(p->name);
			if(p->icon_file_name) 	
				free(p->icon_file_name);
			if(p->popup_list)	
				free(p->popup_list);
			
			os = p->os_data;
			if(os)
			{
				for(port = os->ports; port; port = os->ports)
				{
					if(port->name)
						free(port->name);
					os->ports = port->next;
					
					free(port);
				}
				free(os);
			}
			
			if(p->link)
			{
				g_list_foreach(p->link, (GFunc)page_object_link_destroy, p);
				g_list_free(p->link);
			}
			
			if(prev == NULL)
				np->page_objects = p->next;
			else
				prev->next = p->next;
			free(p);

			return;
		}
		prev = p;
	}
}

/*
 * delete all highlighted page_objects on a net_page (np can be NULL == use current page)
 */
void net_page_delete_page_object(net_page *np)
{
	page_object *po;
	page_object *next;
	
	if(np==0)
		np = get_current_net_page();
	
	for(po = np->plink; po; po = next)
	{
		next = po->plink;
		page_object_delete(np,po);
	}
	np->plink = NULL;
}			

/*
 * Save page_object notes and remove the notes window
 */
void page_object_hide_notes(GtkWidget *w,page_object *po)
{
	char *new = gtk_editable_get_chars(GTK_EDITABLE(po->notes_widget), 0, -1);
	
	if(po->notes == NULL || strcmp(po->notes,new))
	{
		if(po->notes)
			g_free(po->notes);
		po->notes = new;
	}
	else
	{
		g_free(new);
	}
	
	gtk_widget_destroy(po->notes_window);
}

/*
 * Display the notes window and add any text in it that is stored in the page object
 */
void page_object_show_notes(GtkWidget *w, gpointer data)
{
	net_page *np = get_current_net_page();
	page_object *po = np->plink;
	
	if(po)
	{
		GtkWidget *button;
		GtkWidget *vbox;
		GtkWidget *window;
		GtkWidget *edit;
		char buf[256];
		
		strcpy(buf, "Notes for ");
		strcat(buf, page_object_get_name(po));
		
		window = gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_window_set_title(GTK_WINDOW(window),buf);
		gtk_widget_set_usize(window,200,200);
		gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
		
		vbox = gtk_vbox_new(FALSE, 5);
		
		edit = gtk_text_new(NULL,NULL);
		gtk_text_set_editable(GTK_TEXT(edit),TRUE);
		gtk_box_pack_start(GTK_BOX(vbox), edit, TRUE, TRUE, 0);
		
		button = gtk_button_new_with_label("close");
		gtk_signal_connect(GTK_OBJECT(button), "clicked", page_object_hide_notes, po);
		gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 5);
		
		gtk_container_add(GTK_CONTAINER(window),vbox);
		
		if(po->notes)
			gtk_text_insert(GTK_TEXT(edit),NULL,NULL,NULL,po->notes,-1);
		
		po->notes_widget = edit;
		po->notes_window = window;
		
		gtk_widget_grab_focus(edit);
		gtk_widget_show(vbox);
		gtk_widget_show(button);
		gtk_widget_show(edit);
		gtk_widget_show(window);
	}
}

/*
 * Save page_object info and remove the info window
 */
void page_object_hide_info(GtkWidget *w,page_object *po)
{
	gtk_widget_destroy(po->info_window);
}

/*
 * Display the Info about the page object
 */
void page_object_show_info(GtkWidget *w, gpointer data)
{
	net_page *np = get_current_net_page();
	page_object *po = np->plink;
	os_port_entry *port;
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *window;
	GtkWidget *label;
	GtkWidget *clist;
	GtkWidget *scrolled_window;
	char buf[256];
	char *clist_titles[] = { "Service", "Port", "Protocol", "Service Version" };
	char *bufs[ sizeof(clist_titles)/sizeof(char *) ];
	char port_buf[20];
	int i = 0;
	
	if(po)
	{
		strcpy(buf, "Info for ");
		strcat(buf, page_object_get_name(po));
		
		window = gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_window_set_title(GTK_WINDOW(window),buf);
//		gtk_widget_set_usize(window,200,200);
		gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
		
		vbox = gtk_vbox_new(FALSE, 5);

/* Operating system */
		hbox = gtk_hbox_new(FALSE, 5);
		gtk_widget_show(hbox);
		label = gtk_label_new("Operating System: ");
		gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
		gtk_widget_show(label);
		if(po->os_data)
			sprintf(buf, "%s", po->os_data->os);
		else
			sprintf(buf, "Not known");
		label = gtk_label_new(buf);
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
		gtk_widget_show(hbox);

/* IP Address */		
		hbox = gtk_hbox_new(FALSE, 5);
		gtk_widget_show(hbox);
		label = gtk_label_new("IP Address: ");
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
		label = gtk_label_new( ip2str(po->ip) );		
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);

/* DNS Address */
		hbox = gtk_hbox_new(FALSE, 5);
		gtk_widget_show(hbox);
		label = gtk_label_new("DNS Address: ");
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
		if(po->name)
			label = gtk_label_new( po->name );
		else
			label = gtk_label_new( "Not Known" );
		gtk_widget_show(label);
		gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
		gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);

/* Services running */
		clist = gtk_clist_new_with_titles(sizeof(clist_titles)/sizeof(char*), clist_titles);
		gtk_clist_set_selection_mode(GTK_CLIST(clist),GTK_SELECTION_SINGLE);
		gtk_clist_set_shadow_type(GTK_CLIST(clist), GTK_SHADOW_ETCHED_OUT); 
				
		if(po->os_data)
		{
			for(port = po->os_data->ports; port; port = port->next)
			{
				bufs[0] = port->name;
				
				sprintf(port_buf, "%d", port->port);
				bufs[1] = port_buf;
				
				bufs[2] = ( port->protocol == 6 ? "TCP":"UDP");
				if(port->version)
					bufs[3] = port->version;
				else
					bufs[3] = "";
					
				gtk_clist_prepend(GTK_CLIST(clist), bufs);
				i++;
			}
		}

		if(i == 0)
		{
			gtk_box_pack_start(GTK_BOX(vbox), clist, FALSE, FALSE, 5);
			/*
			 * dont show the clist, cause nothing is in it
			 */
		}
		else
		{
			scrolled_window = gtk_scrolled_window_new(NULL,NULL);
			gtk_container_add(GTK_CONTAINER(scrolled_window), clist);
			gtk_widget_set_usize(scrolled_window,600,350);
			gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, FALSE, FALSE, 5);
			gtk_widget_show(scrolled_window);
			gtk_widget_show(clist);
		}
		gtk_clist_columns_autosize(GTK_CLIST(clist));
		
		button = gtk_button_new_with_label("close");
		gtk_signal_connect(GTK_OBJECT(button), "clicked", page_object_hide_info, po);
		gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 5);
		
		gtk_container_add(GTK_CONTAINER(window),vbox);
		
		po->info_window = window;
		
		gtk_widget_show(button);
		gtk_widget_show(vbox);
		gtk_widget_show(window);
	}
}


void page_object_service_callback(GtkWidget *widget, gpointer data)
{
	int port = (u32)data >> 8;
	int protocol = (u32)data & 0xFF;

	// there can only be one page_object selected that can show the service list,
	// so i use this quick and dirty way to get at the IP address :)
	service_callback( (get_current_net_page())->plink->ip,port, protocol);
}

int net_page_count_plink(net_page *np)
{
	page_object *po;
	int i = 0;
	
	if(is_valid_net_page(np))
	{
		for(po = np->plink; po; po = po->plink, i++)
			;
	}
	
	return(i);
}

void delete_option_callback(net_page *np, page_object *po, GtkWidget *menu, void *data)
{
	GtkWidget *item;
	
	item = gtk_menu_item_new(); // put in a seperator
	gtk_menu_append(GTK_MENU(menu), item);
	gtk_widget_show(item);
	
	item = gtk_menu_item_new_with_label("Delete");
	gtk_signal_connect(GTK_OBJECT(item), "activate",
	                   GTK_SIGNAL_FUNC(do_delete_page_object),
	                   (gpointer)NULL);
	
	gtk_menu_append(GTK_MENU(menu), item);
	gtk_widget_show(item);
}

void notes_option_callback(net_page *np, page_object *po, GtkWidget *menu, void *data)
{
	GtkWidget *item;
	
	/*
	 * dont add anything into the menu if there are more than 1 page_objects
	 * selected (highlighted)
	 */
	if(net_page_count_plink(np) == 1)
	{
		item = gtk_menu_item_new_with_label("Notes");
		gtk_signal_connect(GTK_OBJECT(item), "activate",
		                   GTK_SIGNAL_FUNC(page_object_show_notes),
		                   (gpointer)NULL);
		
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);
	}
}

void info_option_callback(net_page *np, page_object *po, GtkWidget *menu, void *data)
{
	GtkWidget *item;
	
	/*
	 * dont add anything into the menu if there are more than 1 page_objects
	 * selected (highlighted)
	 */
	if(net_page_count_plink(np) == 1)
	{
		item = gtk_menu_item_new_with_label("Detailed Info");
		gtk_signal_connect(GTK_OBJECT(item), "activate",
		                   GTK_SIGNAL_FUNC(page_object_show_info),
		                   (gpointer)NULL);
		
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);
	}
}

void map_option_callback(net_page *np, page_object *po, GtkWidget *menu, void *data)
{
	GtkWidget *item;
	char *name = "Map";
	
	if( net_page_count_plink(np) > 1 || po->link != NULL)
		name = "ReMap";

	item = gtk_menu_item_new_with_label(name);
	gtk_signal_connect(GTK_OBJECT(item), "activate",
	                   GTK_SIGNAL_FUNC(net_page_map_plink),
	                   (gpointer)np);
	
	gtk_menu_append(GTK_MENU(menu), item);
	gtk_widget_show(item);
}

void osscan_option_callback(net_page *np, page_object *po, GtkWidget *menu, void *data)
{
	GtkWidget *item;
	char *name = "OS Scan";
	
	if( net_page_count_plink(np) > 1 || po->os_data != NULL)
		name = "ReOS Scan";

	item = gtk_menu_item_new_with_label(name);
	gtk_signal_connect(GTK_OBJECT(item), "activate",
	                   GTK_SIGNAL_FUNC(net_page_osscan_plink),
	                   (gpointer)np);
	
	gtk_menu_append(GTK_MENU(menu), item);
	gtk_widget_show(item);
}

void probe_option_callback(net_page *np, page_object *po, GtkWidget *menu, void *data)
{
	GtkWidget *item;
	char *name = "Scan port versions";
	
	if( net_page_count_plink(np) > 1 || po->os_data != NULL)
		name = "ReScan port versions";

	item = gtk_menu_item_new_with_label(name);
	gtk_signal_connect(GTK_OBJECT(item), "activate",
	                   GTK_SIGNAL_FUNC(net_page_probe_plink),
	                   (gpointer)np);
	
	gtk_menu_append(GTK_MENU(menu), item);
	gtk_widget_show(item);
}

void dns_option_callback(net_page *np, page_object *po, GtkWidget *menu, void *data)
{
	GtkWidget *item;
	char *name = "Reverse DNS";
	
	if( net_page_count_plink(np) > 1 || po->name != NULL)
		name = "ReReverse DNS";

	item = gtk_menu_item_new_with_label(name);
	gtk_signal_connect(GTK_OBJECT(item), "activate",
	                   GTK_SIGNAL_FUNC(net_page_dns_plink),
	                   (gpointer)np);
	
	gtk_menu_append(GTK_MENU(menu), item);
	gtk_widget_show(item);
}


void page_object_option_list_addall(page_object_option *poo, poo_info *pooi)
{
	if(poo && poo->callback && pooi)
		poo->callback(pooi->np, pooi->po, pooi->menu, poo->data);
}

/*
 * Shows an option list for things to do on a oage_object
 */
void page_object_show_option_list(net_page *np, page_object *po, GdkEventButton *eb)
{
	static GtkWidget *pagemenu = NULL;
	GtkWidget *item, *os_seperator_item;
	os_port_entry *port;
	char buf[100];
	poo_info pooi;
	int did_something = FALSE;
	struct servent *service;
	
	if(pagemenu)
		gtk_widget_destroy(pagemenu);
	
	pagemenu = gtk_menu_new();
	
/* start adding all of the other menues here */
	pooi.np = np;
	pooi.po = po;
	pooi.menu = pagemenu;
	g_list_foreach(page_object_option_list, (GFunc)page_object_option_list_addall, &pooi);

	os_seperator_item = gtk_menu_item_new();
	gtk_menu_append(GTK_MENU(pagemenu), os_seperator_item);
	
	if( net_page_count_plink(np) == 1 )
	{
		if(po->os_data)
		{
			for(port = po->os_data->ports; port; port = port->next)
			{
				service = getservbyport(htons(port->port), (port->protocol == 6 ? "tcp":"udp"));
				if(service)
					snprintf(buf, sizeof(buf), "%s", service->s_name);
				else
				{
					char *serv = get_service(port->port, port->protocol);
					if(serv && serv[0] != '\0')
						snprintf(buf, sizeof(buf), "%s", serv);
					else
						snprintf(buf, sizeof(buf), "Port %d", port->port);
				}
				
				if(port->name)
					free(port->name);
					
				port->name = strdup(buf);
					
				item = gtk_menu_item_new_with_label(buf);
				gtk_signal_connect(GTK_OBJECT(item), "activate",
				                   GTK_SIGNAL_FUNC(page_object_service_callback),
				                   (gpointer)(port->port << 8) + port->protocol);
				
				gtk_menu_append(GTK_MENU(pagemenu), item);
				gtk_widget_show(item);
				
				did_something = TRUE;
			}		
		}
	}
	if(did_something)
		gtk_widget_show(os_seperator_item);

	gtk_menu_popup(GTK_MENU(pagemenu), NULL, NULL, NULL, NULL, eb->button, eb->time);
}

void page_object_option_list_add(char *name, page_object_list_callback callback, void *data)
{
	page_object_option *poo;
	
	if( name && callback && (poo = malloc(sizeof(*poo))) )
	{
		page_object_option_list_remove(name);

		poo->name = strdup(name);
		poo->callback = callback;
		poo->data = data;
		
		page_object_option_list = g_list_append(page_object_option_list, poo);
	}
}

void page_object_option_list_remove(char *name)
{
	GList *gl;
	page_object_option *poo;
	
	for(gl = page_object_option_list; gl; gl = gl->next)
	{
		poo = gl->data;
		if( !strcmp(poo->name, name) )
		{
			page_object_option_list = g_list_remove(page_object_option_list, poo);
			free(poo->name);
			free(poo);
			break;
		}
	}
}

void init_net_page_stuff(void)
{
	/* configure the popup list things */
	page_object_option_list_add("notes", 
	                            (page_object_list_callback)notes_option_callback, 
	                            NULL);
	page_object_option_list_add("info",
	                            (page_object_list_callback)info_option_callback, 
	                            NULL);
	page_object_option_list_add("osscan",
	                            (page_object_list_callback)osscan_option_callback, 
	                            NULL);
	page_object_option_list_add("probe",
	                            (page_object_list_callback)probe_option_callback, 
	                            NULL);
	page_object_option_list_add("map",
	                            (page_object_list_callback)map_option_callback, 
	                            NULL);
	page_object_option_list_add("dns",
	                            (page_object_list_callback)dns_option_callback, 
	                            NULL);
	page_object_option_list_add("delete", 
	                            (page_object_list_callback)delete_option_callback, 
	                            NULL);
}
/*
 * event dispatcher for the page_object
 */
int page_object_event_dispatch(GtkObject *widget, GdkEvent *event, page_object *po)
{
	double x,y;
	GdkEventMotion *em = (GdkEventMotion *)event;
	GdkEventButton *eb = (GdkEventButton *)event;
	page_object *pl;
	net_page *np;
	
	switch( event->type )
	{
		case GDK_BUTTON_PRESS:
			if(eb->button == 1)
			{
				mouse_down_x = eb->x;
				mouse_down_y = eb->y;

				gnome_canvas_item_raise_to_top(GNOME_CANVAS_ITEM(po->group));

				po->flags |= PAGE_OBJECT_GRAB;
				
				if(eb->state & GDK_CONTROL_MASK)
				{
				/* trying to select multiple objects */
					if(po->flags & PAGE_OBJECT_HIGHLIGHT)
					{
					/* selected one that is already highlighted, unhighlight it */
						page_object_highlight_unhighlight(po);
						po->flags &= ~PAGE_OBJECT_GRAB;
					}
					else
					{
					/* it is not highlighted, so highlight it */
						page_object_highlight_highlight(po);
					}
				}
				else
				{
				/* not selecting multiple objects */
					if(!(po->flags & PAGE_OBJECT_HIGHLIGHT))
					{
					/* the object is not highlighted, unhighlight all and then highlight the one selected */
						np = get_current_net_page();
						net_page_highlight_all(np, FALSE);
						page_object_highlight_highlight(po);
						net_page_unset_plink();
					}
					page_object_highlight_highlight(po);
				}
				
				net_page_set_plink();
			}
			else
			{
				if(eb->button == 3)
				{
					if(eb->state & GDK_CONTROL_MASK)
					{
						np = get_current_net_page();
						page_object_highlight_highlight(po);
						net_page_set_plink();
//						while(gtk_events_pending())
//							gtk_main_iteration();
						page_object_show_option_list(np,po,eb);
					}
					else
					{
						np = get_current_net_page();
						if(!(po->flags & PAGE_OBJECT_HIGHLIGHT))
						{
							net_page_highlight_all(np, FALSE);
							page_object_highlight_highlight(po);
							net_page_set_plink();
						}
//						while(gtk_events_pending())
//							gtk_main_iteration();
						page_object_show_option_list(np,po,eb);
					}
					po->flags &= ~PAGE_OBJECT_GRAB;
				}
			}
			break;
			
		case GDK_BUTTON_RELEASE:
			po->flags &= ~PAGE_OBJECT_GRAB;
			if(!options_move_stuff_live)
			{
				page_object *po;
				net_page *np = get_current_net_page();
				for(po = np->plink; po; po = po->plink)
					page_object_links_move(po->link, po);
			}
			break;
		
		case GDK_MOTION_NOTIFY:
			if(po->flags & PAGE_OBJECT_GRAB)
			{
				double xx, yy;
				
//				printf("X = %f\nY = %f\n", em->x, em->y);

				np = get_current_net_page();
				
				if(em->x >= 0.0 && em->x <= CANVAS_FIXED_WIDTH) 
				{
					x = (double)em->x - mouse_down_x;
					mouse_down_x = (double)em->x;
				}
				else
					x = 0;
					
				if(em->y >= 0.0 && em->y <= CANVAS_FIXED_HEIGHT) 
				{
					y = (double)em->y - mouse_down_y;
					mouse_down_y = (double)em->y;
				}
				else
					y = 0;
					
				for(pl = np->plink; pl; pl = pl->plink)
				{
					xx = x;
					yy = y;
					
					if( ((pl->x1 + x) >= 0) && ((pl->x2 + x) <= CANVAS_FIXED_WIDTH) )
						pl->x += x;
					else
						xx = 0;
						
					if( ((pl->y1 + y) >= 0) && ((pl->y2 + y) <= CANVAS_FIXED_HEIGHT) )
						pl->y += y;
					else
						yy = 0;	
					
					gnome_canvas_item_move(GNOME_CANVAS_ITEM(pl->group), xx, yy);
					gnome_canvas_item_get_bounds(GNOME_CANVAS_ITEM(pl->group), 
					                             &pl->x1, &pl->y1, &pl->x2, &pl->y2);
					if(options_move_stuff_live)
						page_object_links_move(pl->link, pl);

	//				printf("x=%f y=%f\n",pl->x, pl->y);
				}
			}
			break;
		
		default:
			break;
	}
	return(0);
}	

/*
 * event function for the canvas (tidbit: the canvas event is called before the page items)
 */
int net_page_event_dispatch(GtkObject *widget, GdkEvent *event, net_page *np)
{
	GdkEventButton *eb = (GdkEventButton *)event;
	GdkEventMotion *em = (GdkEventMotion *)event;
	page_object *po;

	switch( event->type )
	{
		case GDK_BUTTON_PRESS:
			if(eb->button == 1)
			{
				if( !(eb->state & GDK_CONTROL_MASK) )
				{
					if(NULL == gnome_canvas_get_item_at(GNOME_CANVAS(np->canvas),eb->x,eb->y) )
					{
					/* we just hit the page to get rid of the highlight */
						net_page_highlight_all(np, FALSE);
						net_page_unset_plink();
						np->button_state |= GDK_BUTTON_PRESS;
						np->select_x = eb->x;
						np->select_y = eb->y;
					}
				}
			}
			else
			{
				if(NULL == gnome_canvas_get_item_at(GNOME_CANVAS(np->canvas),eb->x,eb->y) )
				{
					if (eb->button == 3)
					{
						GtkWidget *pagemenu = gtk_item_factory_get_widget(main_window->itemf, "/Viewspace");
						gtk_menu_popup(GTK_MENU(pagemenu), NULL, NULL, NULL, NULL, eb->button, eb->time);
					}
				}
			}
			break;
			
		case GDK_BUTTON_RELEASE:
			np->button_state &= ~GDK_BUTTON_PRESS;
			if(np->select_rectangle)
			{
				gtk_object_destroy(GTK_OBJECT(np->select_rectangle));
				np->select_rectangle = NULL;
			}
			break;
		
		case GDK_MOTION_NOTIFY:
			if(np->button_state & GDK_BUTTON_PRESS)
			{
				double x,y,x1,x2,y1,y2;
				
				if(np->select_rectangle == NULL)
				{
					np->select_rectangle = gnome_canvas_item_new(	gnome_canvas_root(GNOME_CANVAS(np->canvas)),
											GNOME_TYPE_CANVAS_RECT,
											"outline_color", "black",
											NULL);
					gnome_canvas_item_move(  GNOME_CANVAS_ITEM(np->select_rectangle), 
								(double)np->select_x,
								(double)np->select_y);
				}
				x = em->x - np->select_x;
				y = em->y - np->select_y;
				
				x1 = (double)(x < 0 ? x : 0 );
				y1 = (double)(y > 0 ? 0 : y);
				x2 = (double)(x < 0 ? 0 : x );
				y2 = (double)(y < 0 ? 0 : y);
								
				gnome_canvas_item_set(  GNOME_CANVAS_ITEM(np->select_rectangle), 
							"x1", x1,
							"y1", y1,
							"x2", x2,
							"y2", y2,
							NULL); 
				x1 += np->select_x;
				y1 += np->select_y;
				x2 += np->select_x;
				y2 += np->select_y;
							
//				printf("x1=%f y1=%f x2=%f y2=%f\n",x1,y1,x2,y2);
				 
				for(po = np->page_objects; po; po = po->next)
				{
					int pox = po->x;
					int poy = po->y;
					if( (pox > x1) && (pox < x2) && (poy > y1) && (poy < y2) )
					{
						if( !(po->flags & PAGE_OBJECT_HIGHLIGHT) )
							page_object_highlight_highlight(po);
					}
					else
					{
						if( (po->flags & PAGE_OBJECT_HIGHLIGHT) )
							page_object_highlight_unhighlight(po);
					}
				}
				net_page_set_plink();
			
			}
			break;

		default:
			break;
	}
	return(0);
}	


